#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ilp - a tag based file indexer
#
# Author: slowpoke <mail+git@slowpoke.io>
#
# This program is Free Software under the non-terms
# of the Anti-License. Do whatever the fuck you want.

"""An immutable set implementation."""

from collections import Set


class funcset(Set):

    """Immutable set for pure functional programming.

    funcset implements some methods of a regular, mutable set (add, discard,
    remove), but instead of mutating itself, it will always return either a new
    instance of itself with the changes, or itself if nothing changed.

    """

    def __init__(self, it=tuple()):
        self.__storage = frozenset(it)

    @classmethod
    def new(cls, it=tuple()):
        return cls(it)

    def __contains__(self, x):
        return x in self.__storage

    def __iter__(self):
        return iter(self.__storage)

    def __len__(self):
        return len(self.__storage)

    def __repr__(self):
        return "{" + ", ".join(map(str, self)) + "}"

    def add(self, item):
        if item in self:
            return self
        return self | (item,)

    def discard(self, item):
        if item not in self:
            return self
        return self - (item,)

    def remove(self, item):
        if item not in self:
            raise KeyError(item)
        return self.discard(item)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ilp - a tag based file indexer
#
# Author: slowpoke <mail+git@slowpoke.io>
#
# This program is Free Software under the non-terms
# of the Anti-License. Do whatever the fuck you want.

"""An immutable dictionary implementation."""

from collections import Mapping


class funcmap(Mapping):

    """Immutable dict for pure functional programming.

    This works basically like a regular dictionary, but doesn't mutate its
    state. All methods on funcmap which would normally mutate something return
    a new instance with the applied changes instead.

    """

    def __init__(self, it_or_map=tuple(), **kwargs):
        # NOTE: if a key is present in both it_or_map
        # and kwargs, the latter takes precedence.
        self.__storage = dict(it_or_map, **kwargs)

    @classmethod
    def new(cls, it_or_map=tuple(), **kwargs):
        return cls(it_or_map, **kwargs)

    def __getitem__(self, key):
        return self.__storage[key]

    def __iter__(self):
        return iter(self.__storage)

    def __len__(self):
        return len(self.__storage)

    def __repr__(self):
        return ("{" +
                ", ".join(["{}: {}".format(*map(repr, item))
                           for item in self.items()]) +
                "}")

    def set(self, key, value):

        def it():
            for i in self.items():
                yield i
            yield (key, value)

        return self.new(it())

    def remove(self, key):
        return self.new(
            [(k, v) for k, v in self.items() if k != key])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ilp - a tag based file indexer
#
# Author: slowpoke <mail+git@slowpoke.io>
#
# This program is Free Software under the non-terms
# of the Anti-License. Do whatever the fuck you want.

import os
import shlex

from distutils.util import strtobool

import plac

from .. import database
from ..utils import file
from ..utils import funcset

from ..index import Index


class ILP:

    """A tag based file indexer."""

    commands = (
        "index",
        "forget",
        "tag",
        "untag",
        "deltag",
        "list",
        "info",
        "search",
        "clear")

    def __init__(self):
        home = os.getenv("HOME")
        self._confdir = os.path.join(home, ".ilp")
        if not file.exists(self._confdir):
            os.mkdir(self._confdir)
        self._dbfile = os.path.join(self._confdir, "database")
        self._index = Index()

    def __enter__(self):
        self._database = database.ShelveDB(self._dbfile)
        if "index" in self._database:
            self._index = self._database.retrieve("index")
        else:
            self._index = Index()
            self._database.store("index", self._index)
        return self

    def __exit__(self, etype, exc, tb):
        self._database.update("index", self._index)
        #self._database.close()
        pass

    def index(self,
              recursive: ("recursively add files below a given directory",
                          "flag",
                          "r"),
              *paths: "files or directories to add to the index"):
        for path in paths:
            if file.isdir(path):
                if not recursive:
                    yield "Is a directory: {}".format(path)
                    yield "(To recursively add files to the index, pass -r)"
                else:
                    for subpath in file.walk(path):
                        for output in self.index(False, subpath):
                            yield output
                continue
            full_path = os.path.abspath(path)
            hashstring = None
            try:
                hashstring = file.hash(full_path)
            except Exception as e:
                yield "Couldn't read file: {}".format(full_path, e)
                continue
            self._index = self._index.add_file(full_path, hashstring)
            yield "Adding to the index: {}".format(path)

    def forget(self,
               recursive: ("recursively remove files below a given directory",
                           "flag",
                           "r"),
               *paths: "files or directories to remove from the index"):
        for path in paths:
            if file.isdir(path):
                if not recursive:
                    yield "Is a directory: {}".format(path)
                    yield "(To recursively remove files from the index, pass -r)"
                else:
                    for subpath in file.walk(path):
                        for output in self.forget(False, subpath):
                            yield output
                continue
            full_path = os.path.abspath(path)
            if full_path in self._index.files:
                self._index = self._index.remove_file(full_path)
                yield "Removing from the index: {}".format(path)
            else:
                yield "{} doesn't exist in the index.".format(path)

    def tag(self,
            path: "file or directory to apply tags to",
            recursive: ("recursively add tags to all files below a given directory",
                        "flag",
                        "r"),
            *tags: "tags to add to the file or directory"):
        full_path = os.path.abspath(path)
        if file.isdir(full_path):
            if recursive:
                for subpath in file.walk(path):
                    for output in self.tag(subpath, False, *tags):
                        yield output
            else:
                yield "Is a directory: {}".format(path)
                yield "(To recursively tag files in a directory, pass -r)"
        else:
            for tag_name in tags:
                if not tag_name in self._index.tags:
                    self._index = self._index.add_tag(tag_name)
                    yield "New tag added: {}".format(tag_name)
                if not self._index.file_has_tag(full_path, tag_name):
                    self._index = self._index.tag_file(full_path, tag_name)
                    yield "{} → {}".format(tag_name, path)
                else:
                    yield "{} is already tagging {}".format(tag_name, path)

    def untag(self,
            path: "file or directory to remove tags from",
            recursive: ("recursively remove tags from all files below a given directory",
                        "flag",
                        "r"),
            *tags: "tags to remove from the file or directory"):
        full_path = os.path.abspath(path)
        if file.isdir(full_path):
            if recursive:
                for subpath in file.walk(path):
                    for output in self.untag(subpath, False, *tags):
                        yield output
            else:
                yield "Is a directory: {}".format(path)
                yield "(To recursively untag files in a directory, pass -r)"
        else:
            for tag_name in tags:
                if not tag_name in self._index.tags:
                    yield "Tag doesn't exist: {}".format(tag_name)
                if self._index.file_has_tag(full_path, tag_name):
                    self._index = self._index.untag_file(full_path, tag_name)
                    yield "Removed {} from {}".format(tag_name, path)
                else:
                    yield "{} isn't tagging {}".format(tag_name, path)

    def deltag(self,
               *tag_names: "tags to delete"):
        for tag_name in tag_names:
            if not tag_name in self._index.tags:
                yield "Tag not in the index: {}".format(tag_name)
            else:
                self._index = self._index.remove_tag(tag_name)
                yield "Deleted {}".format(tag_name)

    def list(self,
             what: ("items to list",
                    "positional",
                    None,
                    None,
                    ("files", "tags"),
                    None)):
        if what == "files":
            for path in self._index.files:
                yield path
        elif what == "tags":
            for tag_name in self._index.tags:
                yield tag_name
        else:
            yield "Can't list {}.".format(what)

    def info(self,
             item: "file or tag to get information about",
             tag: ("get information on tags",
                      "flag",
                      "t")):
        # for better code readability and less accidental collisions
        is_tag = tag
        # get all tags tagging the given path
        if not is_tag:
            path = item
            full_path = os.path.abspath(path)
            if not full_path in self._index.files:
                yield "File not in the index: {}".format(path)
            else:
                for tag_name in self._index.tags_of_file(full_path):
                    yield(tag_name)
        # get all paths this tag is tagging
        else:
            tag_name = item
            if not tag_name in self._index.tags:
                yield "Tag not in the index: {}".format(tag_name)
            else:
                tag = self._index.tags.get(tag_name)
                for hashstring in tag:
                    for path in self._index.hashes.get(hashstring):
                        yield(path)

    def search(self,
               query: "the search query"):
        split = shlex.split(query)
        result = self._build_search_result(
            funcset(self._index.hashes), "or", split)
        for hashstring in result:
            try:
                for path in self._index.hashes.get(hashstring):
                    yield path
            except KeyError:
                yield "Unknown hash: {}".format(hashstring)

    def _build_search_result(self, result, operator, tail):
        if len(tail) == 0:
            return result
        else:
            item, tail = tail[0], tail[1:]
            if item in ("and", "or", "not", "xor"):
                operator = item
                return self._build_search_result(result, operator, tail)
            else:
                if item not in self._index.tags:
                    raise KeyError("Tag not in index: {}".format(item))
                operation = {
                    "or": result.__or__,
                    "and": result.__and__,
                    "not": result.__sub__,
                    "xor": result.__xor__}[operator]
                tag = self._index.tags.get(item)
                result = operation(tag)
                return self._build_search_result(result, operator, tail)

    def clear(self,
              yes: ("Confirm deletion",
                    "flag",
                    "y")):
        """Clear the database."""
        if not yes:
            yield "This command purges the database."
            yield "You will LOSE ALL DATA!"
            yield "If you really want this, pass -y to this command."
        else:
            yield "Purging the database."
            self._index = Index()


def main():
    plac.Interpreter.call(ILP, prompt="ilp> ")

if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ilp - a tag based file indexer
#
# Author: slowpoke <mail+git@slowpoke.io>
#
# This program is Free Software under the non-terms
# of the Anti-License. Do whatever the fuck you want.

import os

from .utils import funcmap, funcset


class Index:

    """A collection of files and their associated tags.

    This object is immutable and changes no other state. All methods which
    would cause a state change return a new index object instead, with the
    changes applied.

    The index contains three major data structures:

        - a mapping of files to their content's hashes
        - a mapping of hashes to all filenames they identify
        - a mapping of all tags names to a set of the filenames they tag

    """

    files = property(lambda self: self._files)
    tags = property(lambda self: self._tags)
    hashes = property(lambda self: self._hashes)

    def __init__(self, files=None, tags=None, hashes=None):
        self._files = funcmap() if files is None else funcmap(files)
        self._tags = funcmap() if tags is None else funcmap(tags)
        self._hashes = funcmap() if hashes is None else funcmap(hashes)

    @classmethod
    def new(cls, files=None, tags=None, hashes=None):
        return cls(files, tags, hashes)

    def add_file(self, path, hashstring):
        """Add a new file to the index."""
        if path in self.files and self.files.get(path) != hashstring:
            raise ValueError(
                "Path exists in the database, but has a different hash.")
        new_files = self.files.set(path, hashstring)
        pathlist = self.hashes.get(hashstring, funcset())
        new_hashes = self.hashes.set(
            hashstring,
            pathlist.add(path))
        return self.new(new_files, self.tags, new_hashes)

    def remove_file(self, path):
        """Remove a path from the index.

        NOTE: Even if you remove the last duplicate of a file from the index,
        it will still store the hash, albeit with an empty list of paths
        associated with it.

        """
        if not path in self.files:
            raise KeyError("Path not in the index: {}".format(path))
        hashstring = self.files.get(path)
        new_files = self.files.remove(path)
        pathlist = self.hashes.get(hashstring, funcset())
        new_hashes = self.hashes.set(
            hashstring,
            pathlist.discard(path))
        return self.new(new_files, self.tags, new_hashes)

    def add_tag(self, name):
        """Add a new tag to the index.

        This method will silently return if the tag already exists.

        """
        if name in self.tags:
            raise KeyError("Tag already in the index: {}".format(name))
        return self.new(
            self.files,
            self.tags.set(name, funcset()),
            self.hashes)

    def remove_tag(self, name):
        """Remove a tag from the index."""
        if name not in self.tags:
            raise KeyError("Tag not in the index: {}".format(name))
        return self.new(
            self.files,
            self.tags.remove(name),
            self.hashes)

    def tag_file(self, path, tag_name):
        """Tag a file."""
        if tag_name not in self.tags:
            raise KeyError("Tag not in the index: {}".format(tag_name))
        if path not in self.files:
            raise KeyError("Path not in the index: {}".format(path))
        hashstring = self.files.get(path)
        new_tag = self.tags.get(tag_name).add(hashstring)
        return self.new(
            self.files,
            self.tags.set(tag_name, new_tag),
            self.hashes)

    def untag_file(self, path, tag_name):
        """Untag a file."""
        if tag_name not in self.tags:
            raise KeyError("Tag not in the index: {}".format(tag_name))
        if path not in self.files:
            raise KeyError("Path not in the index: {}".format(path))
        hashstring = self.files.get(path)
        new_tag = self.tags.get(tag_name).discard(hashstring)
        return self.new(
            self.files,
            self.tags.set(tag_name, new_tag),
            self.hashes)

    def file_has_tag(self, path, tag_name):
        """Check if a file is tagged by a tag."""
        if tag_name not in self.tags:
            # if the tag doesn't exist, no path can be tagged by it.
            return False
        tag = self.tags.get(tag_name)
        hashstring = self.files.get(path)
        return hashstring in tag

    def tags_of_file(self, path):
        """Return the names of all tags which are tagging a file."""
        hashstring = self.files.get(path)
        return funcset(tag_name
                       for tag_name, tag in self.tags.items()
                       if hashstring in tag)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ilp - a tag based file indexer
#
# Author: slowpoke <mail+git@slowpoke.io>
#
# This program is Free Software under the non-terms
# of the Anti-License. Do whatever the fuck you want.

"""An interface to files on the disk.

Since this module deals with touching the disk, all functions it provides
should be considered impure unless noted otherwise.

Pure functions in this module:
    - filter (fnmatch.filter, provided for convenience)
    - hash_bytes (a small wrapper function around hashlib's algorithms)

"""

# Since these modules aren't part of the public API of this module, we import
# them under different names to hide them.
import hashlib as _hashlib
import os as _os
import itertools as _itertools

# for convenience
from os.path import exists, isdir

# NOTE: This overwrites builtins.filter in this module. Should it be needed in
# the future, either ``import builtins'' or store it somewhere before importing
# (e.g. ``_filter'').
from fnmatch import filter


def is_readable(path: str) -> bool:
    """Test whether a file is readable."""
    try:
        with open(path):
            pass
    except:  # we consider any error as the file not being readable
        return False
    return True


def read(path: str) -> bytes:
    """Try to read a file from disk.

    Note that this function assumes that the caller has checked for
    preconditions such as existence and readability of the file, or
    is prepared to handle any exceptions that might occur.

    """
    with open(path, "rb") as f:
        return f.read()


def hash_bytes(bytestream: bytes, hashfunc=_hashlib.sha1) -> bytes:
    """Hash a stream of bytes. Defaults to using sha1.

    To use a different hashing algorithm, pass any of the functions from hashlib
    as a second argument (or anything that conforms to the same interface). For
    example, to use md5:

        import hashlib
        hash_bytes(b'some bytes', hashlib.md5)

    This is a pure function.

    """
    return hashfunc(bytestream).digest()


def hash(path: str, hashfunc=_hashlib.sha1) -> bytes:
    """Read and hash a file's contents."""
    contents = read(path)
    return hash_bytes(contents, hashfunc)


def walk(path: str) -> [str]:
    """Get a list of all non-directory files below the given path."""

    def join(dirname, filenames):
        return [_os.path.join(dirname, filename) for filename in filenames]

    # We don't care for the dirnames part of os.walk, which is the middle part
    # of the tuples it generates.
    return _itertools.chain(
        *[join(dirname, filenames)
          for dirname, _, filenames in _os.walk(path)])


def find(path: str, pattern: str) -> [str]:
    """Recursively find all files below path matching the glob pattern."""
    return filter(walk(path), pattern)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ilp - a tag based file indexer
#
# Author: slowpoke <mail+git@slowpoke.io>
#
# This program is Free Software under the non-terms
# of the Anti-License. Do whatever the fuck you want.

from .funcset import funcset
from .funcmap import funcmap
from .functuple import functuple



# ilp - a tag based file indexer
#
# Author: slowpoke <mail+git@slowpoke.io>
#
# This program is Free Software under the non-terms
# of the Anti-License. Do whatever the fuck you want.

from distutils.core import setup

setup(name='ilp',
      description='a tag-based file indexer',
      long_description="""
      ilp is a tool to index, tag, and search through collections of files.
      """,
      version='0.1.1',
      author='slowpoke',
      author_email='mail+python at slowpoke dot io',
      url='https://github.com/proxypoke/ilp',
      requires=['plac'],
      packages=['ilp', 'ilp.cmdline', 'ilp.utils', 'ilp.database'],
      data_files=[
          ("/usr/share/man/man1", ("ilp.1",)),
          ("/usr/share/doc/ilp", ("README.txt",))],
      entry_points={
          'console_scripts': [
              'ilp = ilp.cmdline:main']},
      classifiers=['Environment :: Console',
                   'Development Status :: 4 - Beta',
                   'License :: Public Domain',
                   'Programming Language :: Python :: 3'],
      license='ANTI-LICENSE')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ilp - a tag based file indexer
#
# Author: slowpoke <mail+git@slowpoke.io>
#
# This program is Free Software under the non-terms
# of the Anti-License. Do whatever the fuck you want.

"""A more convenient tuple imlementation for functional programming."""

from collections import Sequence


class functuple(Sequence):

    def __init__(self, it):
        self._storage = tuple(it)

    def __getitem__(self, index):
        try:
            i = int(index)
        except (TypeError, ValueError):
            raise TypeError(
                "indices must be integers, not {}".format(type(index)))
        l = len(self)
        if i >= l or i < -l:
            raise IndexError("index out of range")
        return self._storage[i]

    def __len__(self):
        return len(self._storage)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ilp - a tag based file indexer
#
# Author: slowpoke <mail+git@slowpoke.io>
#
# This program is Free Software under the non-terms
# of the Anti-License. Do whatever the fuck you want.

import shelve
from . import API


class ShelveDB(API):

    """A small wrapper around shelve.

    This database backend is meant mostly for testing and development purposes.

    """

    def __init__(self, path):
        super().__init__()
        self._path = path
        self._shelve = shelve.open(self._path)

    def _put(self, key, value) -> None:
        self._shelve[key] = value

    def _get(self, key) -> "value":
        return self._shelve[key]

    def _del(self, key) -> None:
        del self._shelve[key]

    def __contains__(self, key) -> bool:
        return key in self._shelve



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Kirschwasser - a tag based file indexer
#
# Author: slowpoke <mail+git@slowpoke.io>
#
# This program is Free Software under the non-terms
# of the Anti-License. Do whatever the fuck you want.

"""database - the storage backend API."""

from abc import ABCMeta, abstractmethod


class DatabaseAPI(metaclass=ABCMeta):

    """Abstract metaclass for all database APIs.

    Subclasses must implement _put, _get, _del, and __contains__.

    Note that the first three shouldn't generally check for key existence, the
    user facing methods (store, update, retrieve, remove) do that.

    """

    @abstractmethod
    def _put(self, key, value) -> None:
        return NotImplemented

    @abstractmethod
    def _get(self, key) -> "value":
        return NotImplemented

    @abstractmethod
    def _del(self, key) -> None:
        return NotImplemented

    @abstractmethod
    def __contains__(self, key) -> bool:
        return NotImplemented

    def store(self, key, value):
        """Store an item in the database.

        This method refuses to change existing values. Use update() for that.

        """
        if key in self:
            raise KeyError("'{}' already exists.".format(key))
        self._put(key, value)

    def update(self, key, value, create=False):
        """Change an existing key in the database.

        By default, this method will raise a KeyError if the key doesn't
        already exist. Set create=True if you want it to create a new key.

        """
        if key not in self and not create:
            raise KeyError("'{}' doesn't exist.".format(key))
        else:
            self._put(key, value)

    def retrieve(self, key):
        """Retrieve a key from the database."""
        if key not in self:
            raise KeyError("'{}' doesn't exist.".format(key))
        return self._get(key)

    def remove(self, key):
        """Delete a key from the database."""
        if key not in self:
            raise KeyError("'{}' doesn't exist.".format(key))
        self._del(key)
